/*
 * Copyright (C) 2006, 2007 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "JSBase.h"
#include "JSBasePrivate.h"

#include "APICast.h"
#include "APIShims.h"
#include "OpaqueJSString.h"
#include "SourceCode.h"
#include <interpreter/CallFrame.h>
#include <runtime/InitializeThreading.h>
#include <runtime/Completion.h>
#include <runtime/JSGlobalObject.h>
#include <runtime/JSLock.h>
#include <runtime/JSObject.h>
#include <wtf/text/StringHash.h>

using namespace JSC;

struct EscapeData
{
    EscapeData()
        : output(NULL)
        , outputSize(0)
        , charsEscaped(0)
        , elapsedSeconds(0)
        , error(false)
        , allocated(false)
    {}

    ~EscapeData()
    {
        clear();
    }

    void clear()
    {
        if (allocated) {
            free((void *)output);
            allocated = false;
        }
        output = NULL;
        outputSize = 0;
        charsEscaped = 0;
        elapsedSeconds = 0;
        error = false;
    }

    const char * output;
    unsigned int outputSize;
    unsigned int charsEscaped;
    double elapsedSeconds;
    bool error;     // If true, there was invalid utf8 data
    bool allocated; // If false, output == utf8Text, outputSize == utf8Size, and no free() should be called
};
static bool escapeUTF8(const char * utf8Text, unsigned int utf8Size, EscapeData & outputData);

JSValueRef JSEvaluateInternal(JSContextRef ctx, const UString* script, JSObjectRef thisObject, JSStringRef sourceURL, int startingLineNumber, JSValueRef* exception)
{
    ExecState* exec = toJS(ctx);
    APIEntryShim entryShim(exec);

    JSObject* jsThisObject = toJS(thisObject);

    // evaluate sets "this" to the global object if it is NULL
    JSGlobalObject* globalObject = exec->dynamicGlobalObject();
    SourceCode source = makeSource(*script, sourceURL->ustring(), TextPosition(OrdinalNumber::fromOneBasedInt(startingLineNumber), OrdinalNumber::first()));

    JSValue evaluationException;
    JSValue returnValue = evaluate(globalObject->globalExec(), globalObject->globalScopeChain(), source, jsThisObject, &evaluationException);

    if (evaluationException) {
        if (exception)
            *exception = toRef(exec, evaluationException);
        return 0;
    }

    if (returnValue)
        return toRef(exec, returnValue);

    // happens, for example, when the only statement is an empty (';') statement
    return toRef(exec, jsUndefined());

}

JSValueRef JSEvaluateScript(JSContextRef ctx, JSStringRef script, JSObjectRef thisObject, JSStringRef sourceURL, int startingLineNumber, JSValueRef* exception)
{
    UString scriptData = script->ustring();
    return JSEvaluateInternal(ctx, &scriptData, thisObject, sourceURL, startingLineNumber, exception);
}

JSValueRef JSEvaluateAsciiScript(JSContextRef ctx, const char* script, unsigned int scriptLength, JSObjectRef thisObject, JSStringRef sourceURL, int startingLineNumber, JSValueRef* exception)
{
    EscapeData escapeData;
    if (escapeUTF8(script, scriptLength, escapeData)) {
        script = escapeData.output;
        scriptLength = escapeData.outputSize;
    }
    ASSERT(!escapeData.error);
    UString scriptData(script, scriptLength);
    // printf("Escaped %u chars (%u total bytes) in %u us (%u bytes per ms)\n", escapeData.charsEscaped, scriptLength, (unsigned int)(escapeData.elapsedSeconds * 1000000.0), (unsigned int)(((double)scriptLength / (escapeData.elapsedSeconds * 1000.0))));
    escapeData.clear(); // scriptData made a copy; free any temporaries before parsing
    return JSEvaluateInternal(ctx, &scriptData, thisObject, sourceURL, startingLineNumber, exception);
}

bool JSCheckScriptSyntax(JSContextRef ctx, JSStringRef script, JSStringRef sourceURL, int startingLineNumber, JSValueRef* exception)
{
    ExecState* exec = toJS(ctx);
    APIEntryShim entryShim(exec);

    SourceCode source = makeSource(script->ustring(), sourceURL->ustring(), TextPosition(OrdinalNumber::fromOneBasedInt(startingLineNumber), OrdinalNumber::first()));

    JSValue syntaxException;
    bool isValidSyntax = checkSyntax(exec->dynamicGlobalObject()->globalExec(), source, &syntaxException);

    if (!isValidSyntax) {
        if (exception)
            *exception = toRef(exec, syntaxException);
        return false;
    }

    return true;
}

void JSGarbageCollect(JSContextRef ctx)
{
    // We used to recommend passing NULL as an argument here, which caused the only heap to be collected.
    // As there is no longer a shared heap, the previously recommended usage became a no-op (but the GC
    // will happen when the context group is destroyed).
    // Because the function argument was originally ignored, some clients may pass their released context here,
    // in which case there is a risk of crashing if another thread performs GC on the same heap in between.
    if (!ctx)
        return;

    ExecState* exec = toJS(ctx);
    APIEntryShim entryShim(exec, false);

    exec->globalData().heap.reportAbandonedObjectGraph();
}

void JSReportExtraMemoryCost(JSContextRef ctx, size_t size)
{
    ExecState* exec = toJS(ctx);
    APIEntryShim entryShim(exec);
    exec->globalData().heap.reportExtraMemoryCost(size);
}

static bool escapeUTF8(const char * utf8Text, unsigned int utf8Size, EscapeData & outputData)
{

    double startTime = WTF::currentTime();

    static const int extraRoom = 16 * 1024; // Arbitrary

    // Maximum amount of space a single code point can take in its encoded form
    // The max should be a surrogate pair of the form "\uDEAD\uBEEF"
    static const int encodedMaxLength = 12;

    int inputSize = utf8Size;

    const unsigned char * input = (const unsigned char *) utf8Text;
    bool hasBOM = false;
    if (utf8Size >= 3)
        hasBOM = input[0] == 0xEF && input[1] == 0xBB && input[2] == 0xBF;

    int outputSize = 0;
    char * output = NULL;

    // We don't want to encode the BOM
    int inputIndex = hasBOM ? 3 : 0;
    int outputIndex = 0;
    int memcpyStartIndex = -1;
    bool hasError = false;
    unsigned int charsEscaped = 0;

    while (inputIndex < inputSize) {
        unsigned int ch = input[inputIndex];
        unsigned int n = 0;
        int bytes = 0;

        if (ch <= 0x7F) {
            if (memcpyStartIndex == -1)
                memcpyStartIndex = inputIndex;
            ++inputIndex;
            continue;
        }

        // UTF8 character encountered.

        if (outputSize == 0) {
            outputSize = (int)((float)utf8Size * 1.25f);
            output = (char *)malloc(outputSize + 1);
        } else if (outputIndex + encodedMaxLength > outputSize) {
            outputSize += extraRoom;
            output = (char *)realloc(output, outputSize + 1);
        }

        if (memcpyStartIndex != -1) {
            int len = inputIndex - memcpyStartIndex;
            if ((outputIndex + len + encodedMaxLength) > outputSize) {
                outputSize += len + extraRoom;
                output = (char *)realloc(output, outputSize + 1);
            }
            memcpy(&output[outputIndex], &input[memcpyStartIndex], len);
            outputIndex += len;
            memcpyStartIndex = -1;
        }

        if ((ch & 0xE0) == 0xC0) {
            n = ch & 31;
            bytes = 2;
        } else if ((ch & 0xF0) == 0xE0) {
            n = ch & 15;
            bytes = 3;
        } else if ((ch & 0xF8) == 0xF0) {
            n = ch & 7;
            bytes = 4;
        } else if ((ch & 0xFC) == 0xF8) {
            n = ch & 3;
            bytes = 5;
            hasError = true;
        } else if ((ch & 0xFE) == 0xFC) {
            n = ch & 1;
            bytes = 6;
            hasError = true;
        } else {
            /* not a valid first byte of a UTF-8 sequence */
            n = ch;
            bytes = 1;
            hasError = true;
        }

        unsigned char * buf = (unsigned char *)input + inputIndex + 1;

        for (int j=0; j < bytes - 1; ++j) {
            if (!buf[j] || ((buf[j] & 0xC0) != 0x80)) {
                hasError = true;
                bytes = j;
                break;
            }
            n = (n << 6) | (buf[j] & 0x3F);
        }

        if (hasError) {
            break;
        }

        if (n > 0xFFFF) {
            // Generate a surrogate pair
            outputIndex += sprintf(output + outputIndex, "\\u%04x\\u%04x",
                                   ((n - 0x10000) >> 10) + 0xD800,
                                   ((n - 0x10000) & 0x3FF) + 0xDC00);
        } else {
            outputIndex += sprintf(output + outputIndex, "\\u%04x", n);
        }
        ++charsEscaped;
        inputIndex += bytes;
    }

    if (output && (memcpyStartIndex != -1)) {
        int len = inputIndex - memcpyStartIndex;
        if ((outputIndex + len) > outputSize) {
            outputSize += len;
            output = (char *)realloc(output, outputSize + 1);
        }
        memcpy(&output[outputIndex], &input[memcpyStartIndex], len);
        outputIndex += len;
        memcpyStartIndex = -1;
    }

    if (hasError && output) {
        free(output);
        output = NULL;
        outputSize = 0;
    }

    outputData.allocated = (output != NULL);
    outputData.charsEscaped = charsEscaped;
    if (outputData.allocated) {
        output[outputIndex] = 0;
        outputData.output = output;
        outputData.outputSize = (unsigned int)outputIndex;
    } else {
        // We can use the original text as is, but we still need to skip
        // the BOM, if we had one.
        outputData.output = hasBOM ? utf8Text + 3 : utf8Text;
        outputData.outputSize = hasBOM ? utf8Size - 3 : utf8Size;
    }
    outputData.error = hasError;
    outputData.elapsedSeconds = WTF::currentTime() - startTime;

    return !hasError;
}
